@@ -1,8 +1,8 @@
 /* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
 
-/* Copyright (C) 2018-2022 Hans Petter Jansson
+/* Copyright (C) 2018-2023 Hans Petter Jansson
  *
- * This file is part of Chafa, a program that turns images into character art.
+ * This file is part of Chafa, a program that shows pictures on text terminals.
  *
  * Chafa is free software: you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published
@@ -92,6 +92,14 @@
  * @CHAFA_OPTIMIZATION_ALL: All optimizations enabled.
  **/
 
+/**
+ * ChafaPassthrough:
+ * @CHAFA_PASSTHROUGH_NONE: No passthrough guards will be used.
+ * @CHAFA_PASSTHROUGH_SCREEN: Passthrough guards for GNU Screen will be used.
+ * @CHAFA_PASSTHROUGH_TMUX: Passthrough guards for tmux will be used.
+ * @CHAFA_PASSTHROUGH_MAX: Last supported passthrough mode plus one.
+ **/
+
 /* Private */
 
 void
@@ -937,3 +945,52 @@ chafa_canvas_config_set_fg_only_enabled (ChafaCanvasConfig *config, gboolean fg_
 
     config->fg_only_enabled = fg_only_enabled;
 }
+
+/**
+ * chafa_canvas_config_get_passthrough:
+ * @config: A #ChafaCanvasConfig
+ *
+ * Returns @config's #ChafaPassthrough setting. This defaults to
+ * #CHAFA_PASSTHROUGH_NONE.
+ *
+ * Passthrough is needed to transmit certain escape codes to the outer terminal
+ * when running in an inner terminal environment like tmux. When enabled, this
+ * will happen automatically as needed, dictated by information contained in a
+ * #ChafaTermInfo.
+ *
+ * Returns: The #ChafaPassthrough setting
+ *
+ * Since: 1.14
+ **/
+ChafaPassthrough
+chafa_canvas_config_get_passthrough (const ChafaCanvasConfig *config)
+{
+    g_return_val_if_fail (config != NULL, CHAFA_PASSTHROUGH_NONE);
+    g_return_val_if_fail (config->refs > 0, CHAFA_PASSTHROUGH_NONE);
+
+    return config->passthrough;
+}
+
+/**
+ * chafa_canvas_config_set_passthrough:
+ * @config: A #ChafaCanvasConfig
+ * @passthrough: A #ChafaPassthrough value
+ *
+ * Indicates which passthrough mode to use. This defaults to
+ * #CHAFA_PASSTHROUGH_NONE.
+ *
+ * Passthrough is needed to transmit certain escape codes to the outer terminal
+ * when running in an inner terminal environment like tmux. When enabled, this
+ * will happen automatically as needed, dictated by information contained in a
+ * #ChafaTermInfo.
+ *
+ * Since: 1.14
+ **/
+void
+chafa_canvas_config_set_passthrough (ChafaCanvasConfig *config, ChafaPassthrough passthrough)
+{
+    g_return_if_fail (config != NULL);
+    g_return_if_fail (config->refs > 0);
+
+    config->passthrough = passthrough;
+}
