@@ -1,8 +1,8 @@
 /* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
 
-/* Copyright (C) 2018-2021 Hans Petter Jansson
+/* Copyright (C) 2018-2023 Hans Petter Jansson
  *
- * This file is part of Chafa, a program that turns images into character art.
+ * This file is part of Chafa, a program that shows pictures on text terminals.
  *
  * Chafa is free software: you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published
@@ -61,6 +61,7 @@ typedef enum
     CHAFA_CANVAS_MODE_FGBG_BGFG,
     CHAFA_CANVAS_MODE_FGBG,
     CHAFA_CANVAS_MODE_INDEXED_8,
+    CHAFA_CANVAS_MODE_INDEXED_16_8,
 
     CHAFA_CANVAS_MODE_MAX
 }
@@ -106,6 +107,18 @@ typedef enum
 }
 ChafaOptimizations;
 
+/* Passthrough modes */
+
+typedef enum
+{
+    CHAFA_PASSTHROUGH_NONE,
+    CHAFA_PASSTHROUGH_SCREEN,
+    CHAFA_PASSTHROUGH_TMUX,
+
+    CHAFA_PASSTHROUGH_MAX
+}
+ChafaPassthrough;
+
 /* Canvas config */
 
 typedef struct ChafaCanvasConfig ChafaCanvasConfig;
@@ -209,6 +222,11 @@ gboolean chafa_canvas_config_get_fg_only_enabled (const ChafaCanvasConfig *confi
 CHAFA_AVAILABLE_IN_1_8
 void chafa_canvas_config_set_fg_only_enabled (ChafaCanvasConfig *config, gboolean fg_only_enabled);
 
+CHAFA_AVAILABLE_IN_1_14
+ChafaPassthrough chafa_canvas_config_get_passthrough (const ChafaCanvasConfig *config);
+CHAFA_AVAILABLE_IN_1_14
+void chafa_canvas_config_set_passthrough (ChafaCanvasConfig *config, ChafaPassthrough passthrough);
+
 G_END_DECLS
 
 #endif /* __CHAFA_CANVAS_CONFIG_H__ */
