@@ -1,8 +1,8 @@
 /* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
 
-/* Copyright (C) 2018-2022 Hans Petter Jansson
+/* Copyright (C) 2018-2023 Hans Petter Jansson
  *
- * This file is part of Chafa, a program that turns images into character art.
+ * This file is part of Chafa, a program that shows pictures on text terminals.
  *
  * Chafa is free software: you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published
@@ -1245,6 +1245,152 @@ find_best_solid_char (ChafaCanvas *canvas)
     return best_char;
 }
 
+static void
+destroy_pixel_canvas (ChafaCanvas *canvas)
+{
+    if (canvas->pixel_canvas)
+    {
+        if (canvas->config.pixel_mode == CHAFA_PIXEL_MODE_SIXELS)
+            chafa_sixel_canvas_destroy (canvas->pixel_canvas);
+        else if (canvas->config.pixel_mode == CHAFA_PIXEL_MODE_KITTY)
+            chafa_kitty_canvas_destroy (canvas->pixel_canvas);
+        else if (canvas->config.pixel_mode == CHAFA_PIXEL_MODE_ITERM2)
+            chafa_iterm2_canvas_destroy (canvas->pixel_canvas);
+
+        canvas->pixel_canvas = NULL;
+    }
+}
+
+static void
+draw_all_pixels (ChafaCanvas *canvas, ChafaPixelType src_pixel_type,
+                 const guint8 *src_pixels,
+                 gint src_width, gint src_height, gint src_rowstride)
+{
+    ChafaAlign halign = CHAFA_ALIGN_START, valign = CHAFA_ALIGN_START;
+    ChafaTuck tuck = CHAFA_TUCK_STRETCH;
+
+    if (src_width == 0 || src_height == 0)
+        return;
+
+    if (canvas->placement)
+    {
+        halign = chafa_placement_get_halign (canvas->placement);
+        valign = chafa_placement_get_valign (canvas->placement);
+        tuck = chafa_placement_get_tuck (canvas->placement);
+    }
+
+    if (canvas->pixels)
+    {
+        g_free (canvas->pixels);
+        canvas->pixels = NULL;
+    }
+
+    destroy_pixel_canvas (canvas);
+
+    if (canvas->config.pixel_mode == CHAFA_PIXEL_MODE_SYMBOLS)
+    {
+        /* Symbol mode */
+
+        /* FIXME: The allocation can fail if the canvas is ridiculously large.
+         * Since there's no way to report an error from here, we'll silently
+         * skip the update instead.
+         *
+         * We really shouldn't need this much temporary memory in the first place;
+         * it'd be possible to process the image in cell_height strips and hand
+         * each strip off to the update_cells() pass independently. The pipelining
+         * would improve throughput too. */
+
+        canvas->pixels = g_try_new (ChafaPixel, (gsize) canvas->width_pixels * canvas->height_pixels);
+        if (canvas->pixels)
+        {
+            chafa_prepare_pixel_data_for_symbols (&canvas->fg_palette, &canvas->dither,
+                                                  canvas->config.color_space,
+                                                  canvas->config.preprocessing_enabled,
+                                                  canvas->work_factor_int,
+                                                  src_pixel_type,
+                                                  src_pixels,
+                                                  src_width, src_height,
+                                                  src_rowstride,
+                                                  canvas->pixels,
+                                                  canvas->width_pixels, canvas->height_pixels);
+
+            if (canvas->config.alpha_threshold == 0)
+                canvas->have_alpha = FALSE;
+
+            update_cells (canvas);
+            canvas->needs_clear = FALSE;
+
+            g_free (canvas->pixels);
+            canvas->pixels = NULL;
+        }
+        else
+        {
+#if 0
+            g_warning ("ChafaCanvas: Out of memory allocating %ux%u pixels.",
+                       canvas->width_pixels, canvas->height_pixels);
+#endif
+        }
+    }
+    else if (canvas->config.pixel_mode == CHAFA_PIXEL_MODE_SIXELS)
+    {
+        /* Sixel mode */
+
+        canvas->fg_palette.alpha_threshold = canvas->config.alpha_threshold;
+        canvas->pixel_canvas = chafa_sixel_canvas_new (canvas->width_pixels,
+                                                       canvas->height_pixels,
+                                                       canvas->config.color_space,
+                                                       &canvas->fg_palette,
+                                                       &canvas->dither);
+        chafa_sixel_canvas_draw_all_pixels (canvas->pixel_canvas,
+                                            src_pixel_type,
+                                            src_pixels,
+                                            src_width, src_height,
+                                            src_rowstride,
+                                            halign, valign,
+                                            tuck);
+    }
+    else if (canvas->config.pixel_mode == CHAFA_PIXEL_MODE_KITTY)
+    {
+        ChafaColor bg_color;
+
+        chafa_unpack_color (canvas->config.bg_color_packed_rgb, &bg_color);
+        bg_color.ch [3] = canvas->config.alpha_threshold < 1 ? 0x00 : 0xff;
+
+        /* Kitty mode */
+
+        canvas->fg_palette.alpha_threshold = canvas->config.alpha_threshold;
+        canvas->pixel_canvas = chafa_kitty_canvas_new (canvas->width_pixels,
+                                                       canvas->height_pixels);
+
+        if (canvas->pixel_canvas)
+            chafa_kitty_canvas_draw_all_pixels (canvas->pixel_canvas,
+                                                src_pixel_type,
+                                                src_pixels,
+                                                src_width, src_height,
+                                                src_rowstride,
+                                                bg_color,
+                                                halign, valign,
+                                                tuck);
+    }
+    else  /* if (canvas->config.pixel_mode == CHAFA_PIXEL_MODE_ITERM2) */
+    {
+        /* iTerm2 mode */
+
+        canvas->fg_palette.alpha_threshold = canvas->config.alpha_threshold;
+        canvas->pixel_canvas = chafa_iterm2_canvas_new (canvas->width_pixels,
+                                                        canvas->height_pixels);
+
+        if (canvas->pixel_canvas)
+            chafa_iterm2_canvas_draw_all_pixels (canvas->pixel_canvas,
+                                                 src_pixel_type,
+                                                 src_pixels,
+                                                 src_width, src_height,
+                                                 src_rowstride,
+                                                 halign, valign,
+                                                 tuck);
+    }
+}
+
 /**
  * chafa_canvas_new:
  * @config: Configuration to use or %NULL for hardcoded defaults
@@ -1306,6 +1452,7 @@ chafa_canvas_new (const ChafaCanvasConfig *config)
     canvas->work_factor_int = canvas->config.work_factor * 10 + 0.5f;
     canvas->needs_clear = TRUE;
     canvas->have_alpha = FALSE;
+    canvas->placement = NULL;
 
     canvas->consider_inverted = !(canvas->config.fg_only_enabled
                                   || canvas->config.canvas_mode == CHAFA_CANVAS_MODE_FGBG);
@@ -1406,6 +1553,8 @@ chafa_canvas_new_similar (ChafaCanvas *orig)
 
     chafa_dither_copy (&orig->dither, &canvas->dither);
 
+    canvas->placement = NULL;
+
     return canvas;
 }
 
@@ -1427,22 +1576,6 @@ chafa_canvas_ref (ChafaCanvas *canvas)
     g_atomic_int_inc (&canvas->refs);
 }
 
-static void
-destroy_pixel_canvas (ChafaCanvas *canvas)
-{
-    if (canvas->pixel_canvas)
-    {
-        if (canvas->config.pixel_mode == CHAFA_PIXEL_MODE_SIXELS)
-            chafa_sixel_canvas_destroy (canvas->pixel_canvas);
-        else if (canvas->config.pixel_mode == CHAFA_PIXEL_MODE_KITTY)
-            chafa_kitty_canvas_destroy (canvas->pixel_canvas);
-        else if (canvas->config.pixel_mode == CHAFA_PIXEL_MODE_ITERM2)
-            chafa_iterm2_canvas_destroy (canvas->pixel_canvas);
-
-        canvas->pixel_canvas = NULL;
-    }
-}
-
 /**
  * chafa_canvas_unref:
  * @canvas: Canvas to remove a reference from
@@ -1461,6 +1594,8 @@ chafa_canvas_unref (ChafaCanvas *canvas)
 
     if (g_atomic_int_dec_and_test (&canvas->refs))
     {
+        if (canvas->placement)
+            chafa_placement_unref (canvas->placement);
         chafa_canvas_config_deinit (&canvas->config);
         destroy_pixel_canvas (canvas);
         chafa_dither_deinit (&canvas->dither);
@@ -1491,6 +1626,44 @@ chafa_canvas_peek_config (ChafaCanvas *canvas)
     return &canvas->config;
 }
 
+/**
+ * chafa_canvas_set_placement:
+ * @canvas: Canvas to place the placement on
+ * @placement: Placement to place
+ *
+ * Places @placement on @canvas, replacing the latter's content. The placement
+ * will cover the entire canvas.
+ *
+ * The canvas will keep a reference to the placement until it is replaced or the
+ * canvas itself is freed.
+ *
+ * Since: 1.14
+ */
+void
+chafa_canvas_set_placement (ChafaCanvas *canvas, ChafaPlacement *placement)
+{
+    ChafaImage *image;
+    ChafaFrame *frame;
+
+    g_return_if_fail (canvas != NULL);
+    g_return_if_fail (canvas->refs > 0);
+
+    chafa_placement_ref (placement);
+    if (canvas->placement)
+        chafa_placement_unref (canvas->placement);
+    canvas->placement = placement;
+
+    image = placement->image;
+    g_assert (image != NULL);
+
+    frame = image->frame;
+    if (!frame)
+        return;
+
+    draw_all_pixels (canvas, frame->pixel_type, frame->data,
+                     frame->width, frame->height, frame->rowstride);
+}
+
 /**
  * chafa_canvas_draw_all_pixels:
  * @canvas: Canvas whose pixel data to replace
@@ -1517,91 +1690,8 @@ chafa_canvas_draw_all_pixels (ChafaCanvas *canvas, ChafaPixelType src_pixel_type
     g_return_if_fail (src_width >= 0);
     g_return_if_fail (src_height >= 0);
 
-    if (src_width == 0 || src_height == 0)
-        return;
-
-    if (canvas->pixels)
-    {
-        g_free (canvas->pixels);
-        canvas->pixels = NULL;
-    }
-
-    destroy_pixel_canvas (canvas);
-
-    if (canvas->config.pixel_mode == CHAFA_PIXEL_MODE_SYMBOLS)
-    {
-        /* Symbol mode */
-
-        canvas->pixels = g_new (ChafaPixel, canvas->width_pixels * canvas->height_pixels);
-
-        chafa_prepare_pixel_data_for_symbols (&canvas->fg_palette, &canvas->dither,
-                                              canvas->config.color_space,
-                                              canvas->config.preprocessing_enabled,
-                                              canvas->work_factor_int,
-                                              src_pixel_type,
-                                              src_pixels,
-                                              src_width, src_height,
-                                              src_rowstride,
-                                              canvas->pixels,
-                                              canvas->width_pixels, canvas->height_pixels);
-
-        if (canvas->config.alpha_threshold == 0)
-            canvas->have_alpha = FALSE;
-
-        update_cells (canvas);
-        canvas->needs_clear = FALSE;
-
-        g_free (canvas->pixels);
-        canvas->pixels = NULL;
-    }
-    else if (canvas->config.pixel_mode == CHAFA_PIXEL_MODE_SIXELS)
-    {
-        /* Sixel mode */
-
-        canvas->fg_palette.alpha_threshold = canvas->config.alpha_threshold;
-        canvas->pixel_canvas = chafa_sixel_canvas_new (canvas->width_pixels,
-                                                       canvas->height_pixels,
-                                                       canvas->config.color_space,
-                                                       &canvas->fg_palette,
-                                                       &canvas->dither);
-        chafa_sixel_canvas_draw_all_pixels (canvas->pixel_canvas,
-                                            src_pixel_type,
-                                            src_pixels,
-                                            src_width, src_height,
-                                            src_rowstride);
-    }
-    else if (canvas->config.pixel_mode == CHAFA_PIXEL_MODE_KITTY)
-    {
-        ChafaColor bg_color;
-
-        chafa_unpack_color (canvas->config.bg_color_packed_rgb, &bg_color);
-        bg_color.ch [3] = canvas->config.alpha_threshold < 1 ? 0x00 : 0xff;
-
-        /* Kitty mode */
-
-        canvas->fg_palette.alpha_threshold = canvas->config.alpha_threshold;
-        canvas->pixel_canvas = chafa_kitty_canvas_new (canvas->width_pixels,
-                                                       canvas->height_pixels);
-        chafa_kitty_canvas_draw_all_pixels (canvas->pixel_canvas,
-                                            src_pixel_type,
-                                            src_pixels,
-                                            src_width, src_height,
-                                            src_rowstride,
-                                            bg_color);
-    }
-    else  /* if (canvas->config.pixel_mode == CHAFA_PIXEL_MODE_ITERM2) */
-    {
-        /* iTerm2 mode */
-
-        canvas->fg_palette.alpha_threshold = canvas->config.alpha_threshold;
-        canvas->pixel_canvas = chafa_iterm2_canvas_new (canvas->width_pixels,
-                                                        canvas->height_pixels);
-        chafa_iterm2_canvas_draw_all_pixels (canvas->pixel_canvas,
-                                             src_pixel_type,
-                                             src_pixels,
-                                             src_width, src_height,
-                                             src_rowstride);
-    }
+    draw_all_pixels (canvas, src_pixel_type, src_pixels,
+                     src_width, src_height, src_rowstride);
 }
 
 /**
@@ -1623,8 +1713,8 @@ void
 chafa_canvas_set_contents_rgba8 (ChafaCanvas *canvas, const guint8 *src_pixels,
                                  gint src_width, gint src_height, gint src_rowstride)
 {
-    chafa_canvas_draw_all_pixels (canvas, CHAFA_PIXEL_RGBA8_UNASSOCIATED,
-                                  src_pixels, src_width, src_height, src_rowstride);
+    draw_all_pixels (canvas, CHAFA_PIXEL_RGBA8_UNASSOCIATED,
+                     src_pixels, src_width, src_height, src_rowstride);
 }
 
 /**
@@ -1657,7 +1747,7 @@ chafa_canvas_build_ansi (ChafaCanvas *canvas)
  * @term_info: Terminal to format for, or %NULL for fallback
  *
  * Builds a UTF-8 string of terminal control sequences and symbols
- * representing the canvas' current contents. This can e.g. be printed
+ * representing the canvas' current contents. This can be printed
  * to a terminal. The exact choice of escape sequences and symbols,
  * dimensions, etc. is determined by the configuration assigned to
  * @canvas on its creation.
@@ -1687,34 +1777,29 @@ chafa_canvas_print (ChafaCanvas *canvas, ChafaTermInfo *term_info)
         str = chafa_canvas_print_symbols (canvas, term_info);
     }
     else if (canvas->config.pixel_mode == CHAFA_PIXEL_MODE_SIXELS
-             && chafa_term_info_get_seq (term_info, CHAFA_TERM_SEQ_BEGIN_SIXELS))
+             && chafa_term_info_get_seq (term_info, CHAFA_TERM_SEQ_BEGIN_SIXELS)
+             && canvas->pixel_canvas)
     {
-        gchar buf [CHAFA_TERM_SEQ_LENGTH_MAX + 1];
-        gchar *out;
-
         /* Sixel mode */
 
-        out = chafa_term_info_emit_begin_sixels (term_info, buf, 0, 1, 0);
-        *out = '\0';
-        str = g_string_new (buf);
-
-        g_string_append_printf (str, "\"1;1;%d;%d", canvas->width_pixels, canvas->height_pixels);
-        chafa_sixel_canvas_build_ansi (canvas->pixel_canvas, str);
-
-        out = chafa_term_info_emit_end_sixels (term_info, buf);
-        *out = '\0';
-        g_string_append (str, buf);
+        str = g_string_new ("");
+        chafa_sixel_canvas_build_ansi (canvas->pixel_canvas, term_info, str,
+                                       canvas->config.passthrough);
     }
     else if (canvas->config.pixel_mode == CHAFA_PIXEL_MODE_KITTY
-             && chafa_term_info_get_seq (term_info, CHAFA_TERM_SEQ_BEGIN_KITTY_IMMEDIATE_IMAGE_V1))
+             && chafa_term_info_get_seq (term_info, CHAFA_TERM_SEQ_BEGIN_KITTY_IMMEDIATE_IMAGE_V1)
+             && canvas->pixel_canvas)
     {
         /* Kitty mode */
 
         str = g_string_new ("");
         chafa_kitty_canvas_build_ansi (canvas->pixel_canvas, term_info, str,
-                                       canvas->config.width, canvas->config.height);
+                                       canvas->config.width, canvas->config.height,
+                                       canvas->placement ? canvas->placement->id : -1,
+                                       canvas->config.passthrough);
     }
-    else if (canvas->config.pixel_mode == CHAFA_PIXEL_MODE_ITERM2)
+    else if (canvas->config.pixel_mode == CHAFA_PIXEL_MODE_ITERM2
+             && canvas->pixel_canvas)
     {
         /* iTerm2 mode */
 
@@ -1731,6 +1816,113 @@ chafa_canvas_print (ChafaCanvas *canvas, ChafaTermInfo *term_info)
     return str;
 }
 
+/**
+ * chafa_canvas_print_rows:
+ * @canvas: The canvas to generate a printable representation of
+ * @term_info: Terminal to format for, or %NULL for fallback
+ * @array_out: Pointer to storage for resulting array pointer
+ * @array_len_out: Pointer to storage for array's element count, or %NULL
+ *
+ * Builds an array of UTF-8 strings made up of terminal control sequences
+ * and symbols representing the canvas' current contents. These can be
+ * printed to a terminal. The exact choice of escape sequences and symbols,
+ * dimensions, etc. is determined by the configuration assigned to
+ * @canvas on its creation.
+ *
+ * The array will be %NULL-terminated. The element count does not include
+ * the terminator.
+ *
+ * When the canvas' pixel mode is %CHAFA_PIXEL_MODE_SYMBOLS, each element
+ * will hold the contents of exactly one symbol row. There will be no row
+ * separators, newlines or control sequences to reposition the cursor between
+ * rows. Row positioning is left to the caller.
+ *
+ * In other pixel modes, there may be one or more strings, but the splitting
+ * criteria should not be relied on. They must be printed in sequence, exactly
+ * as they appear.
+ *
+ * Since: 1.14
+ **/
+void
+chafa_canvas_print_rows (ChafaCanvas *canvas, ChafaTermInfo *term_info,
+                         GString ***array_out, gint *array_len_out)
+{
+    g_return_if_fail (canvas != NULL);
+    g_return_if_fail (canvas->refs > 0);
+    g_return_if_fail (array_out != NULL);
+
+    if (term_info)
+        chafa_term_info_ref (term_info);
+    else
+        term_info = chafa_term_db_get_fallback_info (chafa_term_db_get_default ());
+
+    if (canvas->config.pixel_mode == CHAFA_PIXEL_MODE_SYMBOLS)
+    {
+        maybe_clear (canvas);
+        chafa_canvas_print_symbol_rows (canvas, term_info, array_out, array_len_out);
+    }
+    else
+    {
+        GString *gs = chafa_canvas_print (canvas, term_info);
+
+        *array_out = g_new (GString *, 2);
+        (*array_out) [0] = gs;
+        (*array_out) [1] = NULL;
+        if (array_len_out)
+            *array_len_out = 1;
+    }
+}
+
+/**
+ * chafa_canvas_print_rows_strv:
+ * @canvas: The canvas to generate a printable representation of
+ * @term_info: Terminal to format for, or %NULL for fallback
+ *
+ * Builds an array of UTF-8 strings made up of terminal control sequences
+ * and symbols representing the canvas' current contents. These can be
+ * printed to a terminal. The exact choice of escape sequences and symbols,
+ * dimensions, etc. is determined by the configuration assigned to
+ * @canvas on its creation.
+ *
+ * The array will be %NULL-terminated and can be freed with g_strfreev().
+ *
+ * When the canvas' pixel mode is %CHAFA_PIXEL_MODE_SYMBOLS, each element
+ * will hold the contents of exactly one symbol row. There will be no row
+ * separators, newlines or control sequences to reposition the cursor between
+ * rows. Row positioning is left to the caller.
+ *
+ * In other pixel modes, there may be one or more strings, but the splitting
+ * criteria should not be relied on. They must be printed in sequence, exactly
+ * as they appear.
+ *
+ * Returns: A %NULL-terminated array of string pointers
+ *
+ * Since: 1.14
+ **/
+gchar **
+chafa_canvas_print_rows_strv (ChafaCanvas *canvas, ChafaTermInfo *term_info)
+{
+    GString **gsa;
+    gint gsa_len;
+    gchar **strv;
+    gint i;
+
+    g_return_val_if_fail (canvas != NULL, NULL);
+    g_return_val_if_fail (canvas->refs > 0, NULL);
+
+    chafa_canvas_print_rows (canvas, term_info, &gsa, &gsa_len);
+    strv = g_new (gchar *, gsa_len + 1);
+
+    for (i = 0; i < gsa_len; i++)
+    {
+        /* FIXME: Switch to g_string_free_and_steal() in 2025 */
+        strv [i] = g_string_free (gsa [i], FALSE);
+    }
+
+    strv [gsa_len] = NULL;
+    return strv;
+}
+
 /**
  * chafa_canvas_get_char_at:
  * @canvas: The canvas to inspect
