@@ -1,8 +1,8 @@
 /* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
 
-/* Copyright (C) 2018-2022 Hans Petter Jansson
+/* Copyright (C) 2018-2023 Hans Petter Jansson
  *
- * This file is part of Chafa, a program that turns images into character art.
+ * This file is part of Chafa, a program that shows pictures on text terminals.
  *
  * Chafa is free software: you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published
@@ -109,6 +109,7 @@ Glyph2;
  * @CHAFA_SYMBOL_TAG_SEXTANT: Sextant 2x3 mosaics.
  * @CHAFA_SYMBOL_TAG_WEDGE: Wedge shapes that align with sextants.
  * @CHAFA_SYMBOL_TAG_LATIN: Latin and Latin-like symbols (superset of ASCII).
+ * @CHAFA_SYMBOL_TAG_IMPORTED: Symbols for which glyphs were imported with chafa_symbol_map_add_glyph ().
  * @CHAFA_SYMBOL_TAG_EXTRA: Symbols not in any other category.
  * @CHAFA_SYMBOL_TAG_BAD: Joint set of ugly and ambiguous characters. Always excluded unless specifically asked for.
  * @CHAFA_SYMBOL_TAG_ALL: Special value meaning all supported symbols.
@@ -278,11 +279,12 @@ glyph_to_bitmap (gint width, gint height,
 
     /* Scale to cell dimensions */
 
-    smol_scale_simple ((SmolPixelType) pixel_format, pixels, width, height, rowstride,
-                       SMOL_PIXEL_RGBA8_PREMULTIPLIED,
+    smol_scale_simple (pixels, (SmolPixelType) pixel_format, width, height, rowstride,
                        (gpointer) scaled_pixels,
+                       SMOL_PIXEL_RGBA8_UNASSOCIATED,  /* FIXME: Premul */
                        CHAFA_SYMBOL_WIDTH_PIXELS, CHAFA_SYMBOL_HEIGHT_PIXELS,
-                       CHAFA_SYMBOL_WIDTH_PIXELS * 4);
+                       CHAFA_SYMBOL_WIDTH_PIXELS * 4,
+                       SMOL_NO_FLAGS);
 
     /* Generate coverage map */
 
@@ -307,11 +309,12 @@ glyph_to_bitmap_wide (gint width, gint height,
 
     /* Scale to cell dimensions */
 
-    smol_scale_simple ((SmolPixelType) pixel_format, pixels, width, height, rowstride,
-                       SMOL_PIXEL_RGBA8_PREMULTIPLIED,
+    smol_scale_simple (pixels, (SmolPixelType) pixel_format, width, height, rowstride,
                        (gpointer) scaled_pixels,
+                       SMOL_PIXEL_RGBA8_UNASSOCIATED,  /* FIXME: Premul */
                        CHAFA_SYMBOL_WIDTH_PIXELS * 2, CHAFA_SYMBOL_HEIGHT_PIXELS,
-                       CHAFA_SYMBOL_WIDTH_PIXELS * 4 * 2);
+                       CHAFA_SYMBOL_WIDTH_PIXELS * 4 * 2,
+                       SMOL_NO_FLAGS);
 
     /* Generate coverage map */
 
@@ -578,7 +581,7 @@ rebuild_symbols (ChafaSymbolMap *symbol_map)
     while (g_hash_table_iter_next (&iter, &key, &value))
     {
         Glyph *glyph = value;
-        ChafaSymbolTags tags = chafa_get_tags_for_char (glyph->c);
+        ChafaSymbolTags tags = chafa_get_tags_for_char (glyph->c) | CHAFA_SYMBOL_TAG_IMPORTED;
 
         if (char_is_selected (symbol_map->selectors, tags, glyph->c))
         {
@@ -605,7 +608,7 @@ rebuild_symbols (ChafaSymbolMap *symbol_map)
     while (g_hash_table_iter_next (&iter, &key, &value))
     {
         Glyph2 *glyph = value;
-        ChafaSymbolTags tags = chafa_get_tags_for_char (glyph->c);
+        ChafaSymbolTags tags = chafa_get_tags_for_char (glyph->c) | CHAFA_SYMBOL_TAG_IMPORTED;
 
         if (char_is_selected (symbol_map->selectors, tags, glyph->c))
         {
@@ -691,7 +694,7 @@ copy_selector_array (GArray *src)
 }
 
 static void
-add_by_tags (ChafaSymbolMap *symbol_map, ChafaSymbolTags tags)
+add_by_tags (GArray *selectors, ChafaSymbolTags tags)
 {
     Selector s = { 0 };
 
@@ -699,11 +702,11 @@ add_by_tags (ChafaSymbolMap *symbol_map, ChafaSymbolTags tags)
     s.additive = TRUE;
     s.tags = tags;
 
-    g_array_append_val (symbol_map->selectors, s);
+    g_array_append_val (selectors, s);
 }
 
 static void
-remove_by_tags (ChafaSymbolMap *symbol_map, ChafaSymbolTags tags)
+remove_by_tags (GArray *selectors, ChafaSymbolTags tags)
 {
     Selector s = { 0 };
 
@@ -711,11 +714,11 @@ remove_by_tags (ChafaSymbolMap *symbol_map, ChafaSymbolTags tags)
     s.additive = FALSE;
     s.tags = tags;
 
-    g_array_append_val (symbol_map->selectors, s);
+    g_array_append_val (selectors, s);
 }
 
 static void
-add_by_range (ChafaSymbolMap *symbol_map, gunichar first, gunichar last)
+add_by_range (GArray *selectors, gunichar first, gunichar last)
 {
     Selector s = { 0 };
 
@@ -724,11 +727,11 @@ add_by_range (ChafaSymbolMap *symbol_map, gunichar first, gunichar last)
     s.first_code_point = first;
     s.last_code_point = last;
 
-    g_array_append_val (symbol_map->selectors, s);
+    g_array_append_val (selectors, s);
 }
 
 static void
-remove_by_range (ChafaSymbolMap *symbol_map, gunichar first, gunichar last)
+remove_by_range (GArray *selectors, gunichar first, gunichar last)
 {
     Selector s = { 0 };
 
@@ -737,7 +740,7 @@ remove_by_range (ChafaSymbolMap *symbol_map, gunichar first, gunichar last)
     s.first_code_point = first;
     s.last_code_point = last;
 
-    g_array_append_val (symbol_map->selectors, s);
+    g_array_append_val (selectors, s);
 }
 
 static gboolean
@@ -828,6 +831,8 @@ parse_symbol_tag (const gchar *name, gint len, SelectorType *sel_type_out,
         { "bad", CHAFA_SYMBOL_TAG_BAD },
         { "legacy", CHAFA_SYMBOL_TAG_LEGACY },
         { "latin", CHAFA_SYMBOL_TAG_LATIN },
+        { "import", CHAFA_SYMBOL_TAG_IMPORTED },
+        { "imported", CHAFA_SYMBOL_TAG_IMPORTED },
 
         { NULL, 0 }
     };
@@ -882,8 +887,9 @@ static gboolean
 parse_selectors (ChafaSymbolMap *symbol_map, const gchar *selectors, GError **error)
 {
     const gchar *p0 = selectors;
-    gboolean is_add = FALSE, is_remove = FALSE;
+    gboolean is_add = FALSE, is_remove = FALSE, do_clear = FALSE;
     gboolean result = FALSE;
+    GArray *selector_array = g_array_new (FALSE, FALSE, sizeof (Selector));
 
     while (*p0)
     {
@@ -913,46 +919,88 @@ parse_selectors (ChafaSymbolMap *symbol_map, const gchar *selectors, GError **er
         if (!*p0)
             break;
 
-        n = strspn (p0, "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.");
-        if (!n)
+        if (!is_add && !is_remove)
+        {
+            do_clear = TRUE;
+            is_add = TRUE;
+        }
+
+        if (*p0 == '[')
+        {
+            gboolean escape = FALSE;
+
+            p0++;
+
+            for ( ; p0 && *p0; p0 = g_utf8_next_char (p0))
+            {
+                gunichar c = g_utf8_get_char (p0);
+
+                if (c == '\\' && !escape)
+                {
+                    escape = TRUE;
+                    continue;
+                }
+
+                if (c == ']' && !escape)
+                    break;
+
+                if (is_add)
+                    add_by_range (selector_array, c, c);
+                else if (is_remove)
+                    remove_by_range (selector_array, c, c);
+
+                escape = FALSE;
+            }
+
+            if (!p0 || *p0 != ']')
+            {
+                g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE,
+                             "Syntax error in symbol selector set.");
+                goto out;
+            }
+
+            n = 1;
+        }
+        else if (!(n = strspn (p0, "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.")))
         {
             g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE,
                          "Syntax error in symbol tag selectors.");
             goto out;
         }
-
-        if (!parse_symbol_tag (p0, n, &sel_type, &sc, &first, &last, error))
+        else if (!parse_symbol_tag (p0, n, &sel_type, &sc, &first, &last, error))
+        {
             goto out;
-
-        p0 += n;
-
-        if (!is_add && !is_remove)
-        {
-            g_array_set_size (symbol_map->selectors, 0);
-            is_add = TRUE;
-        }
-
-        if (sel_type == SELECTOR_TAG)
-        {
-            if (is_add)
-                add_by_tags (symbol_map, sc);
-            else if (is_remove)
-                remove_by_tags (symbol_map, sc);
         }
         else
         {
-            if (is_add)
-                add_by_range (symbol_map, first, last);
-            else if (is_remove)
-                remove_by_range (symbol_map, first, last);
+            if (sel_type == SELECTOR_TAG)
+            {
+                if (is_add)
+                    add_by_tags (selector_array, sc);
+                else if (is_remove)
+                    remove_by_tags (selector_array, sc);
+            }
+            else
+            {
+                if (is_add)
+                    add_by_range (selector_array, first, last);
+                else if (is_remove)
+                    remove_by_range (selector_array, first, last);
+            }
         }
+
+        p0 += n;
     }
 
-    symbol_map->need_rebuild = TRUE;
+    if (do_clear)
+        g_array_set_size (symbol_map->selectors, 0);
+    g_array_append_vals (symbol_map->selectors, selector_array->data, selector_array->len);
 
+    symbol_map->need_rebuild = TRUE;
     result = TRUE;
 
 out:
+    g_array_free (selector_array, TRUE);
     return result;
 }
 
@@ -1526,7 +1574,7 @@ chafa_symbol_map_add_by_tags (ChafaSymbolMap *symbol_map, ChafaSymbolTags tags)
     g_return_if_fail (symbol_map != NULL);
     g_return_if_fail (symbol_map->refs > 0);
 
-    add_by_tags (symbol_map, tags);
+    add_by_tags (symbol_map->selectors, tags);
 
     symbol_map->need_rebuild = TRUE;
 }
@@ -1544,7 +1592,7 @@ chafa_symbol_map_remove_by_tags (ChafaSymbolMap *symbol_map, ChafaSymbolTags tag
     g_return_if_fail (symbol_map != NULL);
     g_return_if_fail (symbol_map->refs > 0);
 
-    remove_by_tags (symbol_map, tags);
+    remove_by_tags (symbol_map->selectors, tags);
 
     symbol_map->need_rebuild = TRUE;
 }
@@ -1566,7 +1614,7 @@ chafa_symbol_map_add_by_range (ChafaSymbolMap *symbol_map, gunichar first, gunic
     g_return_if_fail (symbol_map != NULL);
     g_return_if_fail (symbol_map->refs > 0);
 
-    add_by_range (symbol_map, first, last);
+    add_by_range (symbol_map->selectors, first, last);
 
     symbol_map->need_rebuild = TRUE;
 }
@@ -1588,7 +1636,7 @@ chafa_symbol_map_remove_by_range (ChafaSymbolMap *symbol_map, gunichar first, gu
     g_return_if_fail (symbol_map != NULL);
     g_return_if_fail (symbol_map->refs > 0);
 
-    remove_by_range (symbol_map, first, last);
+    remove_by_range (symbol_map->selectors, first, last);
 
     symbol_map->need_rebuild = TRUE;
 }
@@ -1826,10 +1874,11 @@ chafa_symbol_map_get_glyph (ChafaSymbolMap *symbol_map,
         gpointer temp_pixels = g_malloc (width * CHAFA_SYMBOL_HEIGHT_PIXELS * 4);
 
         /* Convert to desired pixel format */
-        smol_scale_simple (SMOL_PIXEL_ARGB8_PREMULTIPLIED, *pixels_out,
+        smol_scale_simple (*pixels_out, SMOL_PIXEL_ARGB8_UNASSOCIATED,  /* FIXME: Premul */
                            width, height, rowstride,
-                           (SmolPixelType) pixel_format, temp_pixels,
-                           width, height, rowstride);
+                           temp_pixels, (SmolPixelType) pixel_format,
+                           width, height, rowstride,
+                           SMOL_NO_FLAGS);
         g_free (*pixels_out);
         *pixels_out = temp_pixels;
     }
