@@ -1,8 +1,8 @@
 /* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
 
-/* Copyright (C) 2020-2022 Hans Petter Jansson
+/* Copyright (C) 2020-2023 Hans Petter Jansson
  *
- * This file is part of Chafa, a program that turns images into character art.
+ * This file is part of Chafa, a program that shows pictures on text terminals.
  *
  * Chafa is free software: you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published
@@ -67,8 +67,8 @@ static const SeqStr vt220_seqs [] =
     { CHAFA_TERM_SEQ_CURSOR_LEFT_1, "\033[D" },
     { CHAFA_TERM_SEQ_CURSOR_RIGHT, "\033[%1C" },
     { CHAFA_TERM_SEQ_CURSOR_RIGHT_1, "\033[C" },
-    { CHAFA_TERM_SEQ_CURSOR_UP_SCROLL, "\033D" },
-    { CHAFA_TERM_SEQ_CURSOR_DOWN_SCROLL, "\033M" },
+    { CHAFA_TERM_SEQ_CURSOR_UP_SCROLL, "\033M" },
+    { CHAFA_TERM_SEQ_CURSOR_DOWN_SCROLL, "\033D" },
     { CHAFA_TERM_SEQ_INSERT_CELLS, "\033[%1@" },
     { CHAFA_TERM_SEQ_DELETE_CELLS, "\033[%1P" },
     { CHAFA_TERM_SEQ_INSERT_ROWS, "\033[%1L" },
@@ -82,6 +82,13 @@ static const SeqStr vt220_seqs [] =
     { CHAFA_TERM_SEQ_DISABLE_ECHO, "\033[12h" },
     { CHAFA_TERM_SEQ_ENABLE_WRAP, "\033[?7h" },
     { CHAFA_TERM_SEQ_DISABLE_WRAP, "\033[?7l" },
+    { CHAFA_TERM_SEQ_RESET_SCROLLING_ROWS, "\033[r" },
+    { CHAFA_TERM_SEQ_SAVE_CURSOR_POS, "\033[s" },
+    { CHAFA_TERM_SEQ_RESTORE_CURSOR_POS, "\033[u" },
+
+    /* These are actually xterm seqs, but we'll allow it */
+    { CHAFA_TERM_SEQ_ENABLE_ALT_SCREEN, "\033[1049h" },
+    { CHAFA_TERM_SEQ_DISABLE_ALT_SCREEN, "\033[1049l" },
 
     { CHAFA_TERM_SEQ_MAX, NULL }
 };
@@ -99,12 +106,104 @@ static const SeqStr sixel_seqs [] =
     { CHAFA_TERM_SEQ_END_SIXELS, "\033\\" },
     { CHAFA_TERM_SEQ_ENABLE_SIXEL_SCROLLING, "\033[?80l" },
     { CHAFA_TERM_SEQ_DISABLE_SIXEL_SCROLLING, "\033[?80h" },
+    { CHAFA_TERM_SEQ_SET_SIXEL_ADVANCE_DOWN, "\033[?8452l" },
+    { CHAFA_TERM_SEQ_SET_SIXEL_ADVANCE_RIGHT, "\033[?8452h" },
+
+    { CHAFA_TERM_SEQ_MAX, NULL }
+};
+
+static const SeqStr default_color_seqs [] =
+{
+    { CHAFA_TERM_SEQ_RESET_DEFAULT_FG, "\033]110\033\\" },
+    { CHAFA_TERM_SEQ_SET_DEFAULT_FG, "\033]10;rgb:%1/%2/%3\e\\" },
+    { CHAFA_TERM_SEQ_QUERY_DEFAULT_FG, "\033]10;?\033\\" },
+
+    { CHAFA_TERM_SEQ_RESET_DEFAULT_BG, "\033]111\033\\" },
+    { CHAFA_TERM_SEQ_SET_DEFAULT_BG, "\033]11;rgb:%1/%2/%3\e\\" },
+    { CHAFA_TERM_SEQ_QUERY_DEFAULT_BG, "\033]11;?\033\\" },
+
+    { CHAFA_TERM_SEQ_MAX, NULL }
+};
+
+static const SeqStr default_key_seqs [] =
+{
+    { CHAFA_TERM_SEQ_RETURN_KEY, "\x0d" },  /* ASCII CR */
+    { CHAFA_TERM_SEQ_BACKSPACE_KEY, "\x7f" },  /* ASCII DEL */
+    { CHAFA_TERM_SEQ_TAB_KEY, "\x09" },  /* ASCII HT */
+    { CHAFA_TERM_SEQ_TAB_SHIFT_KEY, "\033[Z" },
+    { CHAFA_TERM_SEQ_UP_KEY, "\033[A" },
+    { CHAFA_TERM_SEQ_UP_CTRL_KEY, "\033[1;5A" },
+    { CHAFA_TERM_SEQ_UP_SHIFT_KEY, "\033[1;2A" },
+    { CHAFA_TERM_SEQ_DOWN_KEY, "\033[B" },
+    { CHAFA_TERM_SEQ_DOWN_CTRL_KEY, "\033[1;5B" },
+    { CHAFA_TERM_SEQ_DOWN_SHIFT_KEY, "\033[1;2B" },
+    { CHAFA_TERM_SEQ_LEFT_KEY, "\033[D" },
+    { CHAFA_TERM_SEQ_LEFT_CTRL_KEY, "\033[1;5D" },
+    { CHAFA_TERM_SEQ_LEFT_SHIFT_KEY, "\033[1;2D" },
+    { CHAFA_TERM_SEQ_RIGHT_KEY, "\033[C" },
+    { CHAFA_TERM_SEQ_RIGHT_CTRL_KEY, "\033[1;5C" },
+    { CHAFA_TERM_SEQ_RIGHT_SHIFT_KEY, "\033[1;2C" },
+    { CHAFA_TERM_SEQ_PAGE_UP_KEY, "\033[5~" },
+    { CHAFA_TERM_SEQ_PAGE_UP_CTRL_KEY, "\033[5;5~" },
+    { CHAFA_TERM_SEQ_PAGE_UP_SHIFT_KEY, "\033[5;2~" },
+    { CHAFA_TERM_SEQ_PAGE_DOWN_KEY, "\033[6~" },
+    { CHAFA_TERM_SEQ_PAGE_DOWN_CTRL_KEY, "\033[6;5~" },
+    { CHAFA_TERM_SEQ_PAGE_DOWN_SHIFT_KEY, "\033[6;2~" },
+    { CHAFA_TERM_SEQ_HOME_KEY, "\033[H" },
+    { CHAFA_TERM_SEQ_HOME_CTRL_KEY, "\033[1;5H" },
+    { CHAFA_TERM_SEQ_HOME_SHIFT_KEY, "\033[1;2H" },
+    { CHAFA_TERM_SEQ_END_KEY, "\033[F" },
+    { CHAFA_TERM_SEQ_END_CTRL_KEY, "\033[1;5F" },
+    { CHAFA_TERM_SEQ_END_SHIFT_KEY, "\033[1;2F" },
+    { CHAFA_TERM_SEQ_INSERT_KEY, "\033[2~" },
+    { CHAFA_TERM_SEQ_INSERT_CTRL_KEY, "\033[2;5~" },
+    { CHAFA_TERM_SEQ_INSERT_SHIFT_KEY, "\033[2;2~" },
+    { CHAFA_TERM_SEQ_DELETE_KEY, "\033[3~" },
+    { CHAFA_TERM_SEQ_DELETE_CTRL_KEY, "\033[3;5~" },
+    { CHAFA_TERM_SEQ_DELETE_SHIFT_KEY, "\033[3;2~" },
+    { CHAFA_TERM_SEQ_F1_KEY, "\033OP" },
+    { CHAFA_TERM_SEQ_F1_CTRL_KEY, "\033[1;5P" },
+    { CHAFA_TERM_SEQ_F1_SHIFT_KEY, "\033[1;2P" },
+    { CHAFA_TERM_SEQ_F2_KEY, "\033OQ" },
+    { CHAFA_TERM_SEQ_F2_CTRL_KEY, "\033[1;5Q" },
+    { CHAFA_TERM_SEQ_F2_SHIFT_KEY, "\033[1;2Q" },
+    { CHAFA_TERM_SEQ_F3_KEY, "\033R" },
+    { CHAFA_TERM_SEQ_F3_CTRL_KEY, "\033[1;5R" },
+    { CHAFA_TERM_SEQ_F3_SHIFT_KEY, "\033[1;2R" },
+    { CHAFA_TERM_SEQ_F4_KEY, "\033OS" },
+    { CHAFA_TERM_SEQ_F4_CTRL_KEY, "\033[1;5S" },
+    { CHAFA_TERM_SEQ_F4_SHIFT_KEY, "\033[1;2S" },
+    { CHAFA_TERM_SEQ_F5_KEY, "\033[15~" },
+    { CHAFA_TERM_SEQ_F5_CTRL_KEY, "\033[15;5~" },
+    { CHAFA_TERM_SEQ_F5_SHIFT_KEY, "\033[15;2~" },
+    { CHAFA_TERM_SEQ_F6_KEY, "\033[17~" },
+    { CHAFA_TERM_SEQ_F6_CTRL_KEY, "\033[17;5~" },
+    { CHAFA_TERM_SEQ_F6_SHIFT_KEY, "\033[17;2~" },
+    { CHAFA_TERM_SEQ_F7_KEY, "\033[18~" },
+    { CHAFA_TERM_SEQ_F7_CTRL_KEY, "\033[18;5~" },
+    { CHAFA_TERM_SEQ_F7_SHIFT_KEY, "\033[18;2~" },
+    { CHAFA_TERM_SEQ_F8_KEY, "\033[19~" },
+    { CHAFA_TERM_SEQ_F8_CTRL_KEY, "\033[19;5~" },
+    { CHAFA_TERM_SEQ_F8_SHIFT_KEY, "\033[19;2~" },
+    { CHAFA_TERM_SEQ_F9_KEY, "\033[20~" },
+    { CHAFA_TERM_SEQ_F9_CTRL_KEY, "\033[20;5~" },
+    { CHAFA_TERM_SEQ_F9_SHIFT_KEY, "\033[20;2~" },
+    { CHAFA_TERM_SEQ_F10_KEY, "\033[21~" },
+    { CHAFA_TERM_SEQ_F10_CTRL_KEY, "\033[21;5~" },
+    { CHAFA_TERM_SEQ_F10_SHIFT_KEY, "\033[21;2~" },
+    { CHAFA_TERM_SEQ_F11_KEY, "\033[23~" },
+    { CHAFA_TERM_SEQ_F11_CTRL_KEY, "\033[23;5~" },
+    { CHAFA_TERM_SEQ_F11_SHIFT_KEY, "\033[23;2~" },
+    { CHAFA_TERM_SEQ_F12_KEY, "\033[24~" },
+    { CHAFA_TERM_SEQ_F12_CTRL_KEY, "\033[24;5~" },
+    { CHAFA_TERM_SEQ_F12_SHIFT_KEY, "\033[24;2~" },
 
     { CHAFA_TERM_SEQ_MAX, NULL }
 };
 
 static const SeqStr color_direct_seqs [] =
 {
+    /* ISO 8613-6 */
     { CHAFA_TERM_SEQ_SET_COLOR_FG_DIRECT, "\033[38;2;%1;%2;%3m" },
     { CHAFA_TERM_SEQ_SET_COLOR_BG_DIRECT, "\033[48;2;%1;%2;%3m" },
     { CHAFA_TERM_SEQ_SET_COLOR_FGBG_DIRECT, "\033[38;2;%1;%2;%3;48;2;%4;%5;%6m" },
@@ -136,6 +235,11 @@ static const SeqStr color_8_seqs [] =
     { CHAFA_TERM_SEQ_SET_COLOR_BG_8, "\033[%1m" },
     { CHAFA_TERM_SEQ_SET_COLOR_FGBG_8, "\033[%1;%2m" },
 
+    /* ECMA-48 3rd ed. March 1984 */
+    { CHAFA_TERM_SEQ_RESET_COLOR_FG, "\033[39m" },
+    { CHAFA_TERM_SEQ_RESET_COLOR_BG, "\033[49m" },
+    { CHAFA_TERM_SEQ_RESET_COLOR_FGBG, "\033[39;49m" },
+
     { CHAFA_TERM_SEQ_MAX, NULL }
 };
 
@@ -185,6 +289,7 @@ static const SeqStr *color_fbterm_list [] =
 static const SeqStr kitty_seqs [] =
 {
     { CHAFA_TERM_SEQ_BEGIN_KITTY_IMMEDIATE_IMAGE_V1, "\033_Ga=T,f=%1,s=%2,v=%3,c=%4,r=%5,m=1\033\\" },
+    { CHAFA_TERM_SEQ_BEGIN_KITTY_IMMEDIATE_VIRT_IMAGE_V1, "\033_Ga=T,U=1,q=2,f=%1,s=%2,v=%3,c=%4,r=%5,i=%6,m=1\033\\" },
     { CHAFA_TERM_SEQ_END_KITTY_IMAGE, "\033_Gm=0\033\\" },
     { CHAFA_TERM_SEQ_BEGIN_KITTY_IMAGE_CHUNK, "\033_Gm=1;" },
     { CHAFA_TERM_SEQ_END_KITTY_IMAGE_CHUNK, "\033\\" },
@@ -200,6 +305,22 @@ static const SeqStr iterm2_seqs [] =
     { CHAFA_TERM_SEQ_MAX, NULL }
 };
 
+static const SeqStr tmux_seqs [] =
+{
+    { CHAFA_TERM_SEQ_BEGIN_TMUX_PASSTHROUGH, "\033Ptmux;" },
+    { CHAFA_TERM_SEQ_END_TMUX_PASSTHROUGH, "\033\\" },
+
+    { CHAFA_TERM_SEQ_MAX, NULL }
+};
+
+static const SeqStr screen_seqs [] =
+{
+    { CHAFA_TERM_SEQ_BEGIN_SCREEN_PASSTHROUGH, "\033P" },
+    { CHAFA_TERM_SEQ_END_SCREEN_PASSTHROUGH, "\033\\" },
+
+    { CHAFA_TERM_SEQ_MAX, NULL }
+};
+
 static const SeqStr *fallback_list [] =
 {
     vt220_seqs,
@@ -210,6 +331,8 @@ static const SeqStr *fallback_list [] =
     sixel_seqs,
     kitty_seqs,
     iterm2_seqs,
+    screen_seqs,
+    tmux_seqs,
     NULL
 };
 
@@ -241,6 +364,18 @@ add_seq_list (ChafaTermInfo *ti, const SeqStr **seqlist)
     }
 }
 
+static const gchar *
+getenv_or_blank (gchar **envp, const gchar *key)
+{
+    const gchar *value;
+
+    value = g_environ_getenv (envp, key);
+    if (!value)
+        value = "";
+
+    return value;
+}
+
 static void
 detect_capabilities (ChafaTermInfo *ti, gchar **envp)
 {
@@ -253,39 +388,29 @@ detect_capabilities (ChafaTermInfo *ti, gchar **envp)
     const gchar *tmux;
     const gchar *ctx_backend;
     const gchar *lc_terminal;
+    const gchar *kitty_pid;
+    const gchar *mlterm;
+    const gchar *nvim;
+    const gchar *nvim_tui_enable_true_color;
     gchar *comspec = NULL;
     const SeqStr **color_seq_list = color_256_list;
     const SeqStr *gfx_seqs = NULL;
     const SeqStr *rep_seqs_local = NULL;
+    const SeqStr *inner_seqs = NULL;
 
-    add_seqs (ti, vt220_seqs);
-
-    term = g_environ_getenv (envp, "TERM");
-    if (!term) term = "";
-
-    colorterm = g_environ_getenv (envp, "COLORTERM");
-    if (!colorterm) colorterm = "";
-
-    konsole_version = g_environ_getenv (envp, "KONSOLE_VERSION");
-    if (!konsole_version) konsole_version = "";
-
-    vte_version = g_environ_getenv (envp, "VTE_VERSION");
-    if (!vte_version) vte_version = "";
-
-    term_program = g_environ_getenv (envp, "TERM_PROGRAM");
-    if (!term_program) term_program = "";
-
-    term_name = g_environ_getenv (envp, "TERMINAL_NAME");
-    if (!term_name) term_name = "";
-
-    tmux = g_environ_getenv (envp, "TMUX");
-    if (!tmux) tmux = "";
-
-    ctx_backend = g_environ_getenv (envp, "CTX_BACKEND");
-    if (!ctx_backend) ctx_backend = "";
-
-    lc_terminal = g_environ_getenv (envp, "LC_TERMINAL");
-    if (!lc_terminal) lc_terminal = "";
+    term = getenv_or_blank (envp, "TERM");
+    colorterm = getenv_or_blank (envp, "COLORTERM");
+    konsole_version = getenv_or_blank (envp, "KONSOLE_VERSION");
+    vte_version = getenv_or_blank (envp, "VTE_VERSION");
+    term_program = getenv_or_blank (envp, "TERM_PROGRAM");
+    term_name = getenv_or_blank (envp, "TERMINAL_NAME");
+    tmux = getenv_or_blank (envp, "TMUX");
+    ctx_backend = getenv_or_blank (envp, "CTX_BACKEND");
+    lc_terminal = getenv_or_blank (envp, "LC_TERMINAL");
+    kitty_pid = getenv_or_blank (envp, "KITTY_PID");
+    mlterm = getenv_or_blank (envp, "MLTERM");
+    nvim = getenv_or_blank (envp, "NVIM");
+    nvim_tui_enable_true_color = getenv_or_blank (envp, "NVIM_TUI_ENABLE_TRUE_COLOR");
 
     /* The MS Windows 10 TH2 (v1511+) console supports ANSI escape codes,
      * including AIX and DirectColor sequences. We detect this early and allow
@@ -344,7 +469,8 @@ detect_capabilities (ChafaTermInfo *ti, gchar **envp)
         color_seq_list = color_direct_list;
 
     /* Kitty has a unique graphics protocol */
-    if (!strcmp (term, "xterm-kitty"))
+    if (!strcmp (term, "xterm-kitty")
+        || strlen (kitty_pid) > 0)
         gfx_seqs = kitty_seqs;
 
     /* iTerm2 supports truecolor and has a unique graphics protocol */
@@ -365,6 +491,28 @@ detect_capabilities (ChafaTermInfo *ti, gchar **envp)
         gfx_seqs = sixel_seqs;
     }
 
+    /* Check for Neovim early. It pretends to be xterm-256color, and may or
+     * may not support directcolor. */
+    if (strlen (nvim) > 0)
+    {
+        /* The Neovim terminal defaults to 256 colors unless termguicolors has
+         * been set to true. */
+        color_seq_list = color_256_list;
+
+        /* If COLORTERM was explicitly set to truecolor, honor it. Neovim may do
+         * this when termguicolors has been set to true *and* COLORTERM was
+         * previously set. See Neovim commit d8963c434f01e6a7316 (Nov 26, 2020).
+         *
+         * The user may also set NVIM_TUI_ENABLE_TRUE_COLOR=1 in older Neovim
+         * versions. We'll honor that one blindly, since it's specific and there
+         * seems to be no better option. */
+        if (!g_ascii_strcasecmp (colorterm, "truecolor")
+            || !g_ascii_strcasecmp (nvim_tui_enable_true_color, "1"))
+        {
+            color_seq_list = color_direct_list;
+        }
+    }
+
     /* Apple Terminal sets TERM=xterm-256color, and does not support truecolor */
     if (!g_ascii_strcasecmp (term_program, "Apple_Terminal"))
         color_seq_list = color_256_list;
@@ -375,6 +523,7 @@ detect_capabilities (ChafaTermInfo *ti, gchar **envp)
      * yaft supports sixels and truecolor escape codes, but it remaps cell
      * colors to a 256-color palette. */
     if (!strcmp (term, "mlterm")
+        || strlen (mlterm) > 0
         || !strcmp (term, "yaft")
         || !strcmp (term, "yaft-256color"))
     {
@@ -400,19 +549,28 @@ detect_capabilities (ChafaTermInfo *ti, gchar **envp)
      * like so: screen.xterm-256color */
     if (!strncmp (term, "screen", 6))
     {
-        color_seq_list = color_256_list;
-
         /* 'tmux' also sets TERM=screen, but it supports truecolor codes.
          * You may have to add the following to .tmux.conf to prevent
          * remapping to 256 colors:
          *
          * tmux set-option -ga terminal-overrides ",screen-256color:Tc" */
         if (strlen (tmux) > 0)
+        {
             color_seq_list = color_direct_list;
+            inner_seqs = tmux_seqs;
+        }
+        else
+        {
+            color_seq_list = color_256_list;
+            inner_seqs = screen_seqs;
+        }
 
         /* screen and older tmux do not support REP. Newer tmux does,
          * but there's no reliable way to tell which version we're dealing with. */
         rep_seqs_local = NULL;
+
+        /* Graphics is allowed in screen and tmux, with passthrough. */
+        /* gfx_seqs = NULL; */
     }
 
     /* If TERM is "linux", we're probably on the Linux console, which supports
@@ -430,9 +588,11 @@ detect_capabilities (ChafaTermInfo *ti, gchar **envp)
     if (!strcmp (term, "fbterm"))
         color_seq_list = color_fbterm_list;
 
+    add_seqs (ti, vt220_seqs);
     add_seq_list (ti, color_seq_list);
     add_seqs (ti, gfx_seqs);
     add_seqs (ti, rep_seqs_local);
+    add_seqs (ti, inner_seqs);
 }
 
 static ChafaTermDb *
