@@ -1,8 +1,8 @@
 /* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
 
-/* Copyright (C) 2020-2022 Hans Petter Jansson
+/* Copyright (C) 2020-2023 Hans Petter Jansson
  *
- * This file is part of Chafa, a program that turns images into character art.
+ * This file is part of Chafa, a program that shows pictures on text terminals.
  *
  * Chafa is free software: you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published
@@ -110,6 +110,17 @@ queue_char (PrintCtx *ctx, gchar *out, gunichar c)
     return out;
 }
 
+G_GNUC_WARN_UNUSED_RESULT static gchar *
+reset_fg (PrintCtx *ctx, gchar *out)
+{
+    out = chafa_term_info_emit_reset_color_fg (ctx->term_info, out);
+
+    ctx->cur_fg = CHAFA_PALETTE_INDEX_TRANSPARENT;
+    ctx->cur_fg_direct.ch [3] = 0;
+
+    return out;
+}
+
 G_GNUC_WARN_UNUSED_RESULT static gchar *
 reset_attributes (PrintCtx *ctx, gchar *out)
 {
@@ -657,79 +668,91 @@ prealloc_string (GString *gs, gint n_cells)
     }
 }
 
+G_GNUC_WARN_UNUSED_RESULT static gchar *
+build_ansi_row (PrintCtx *ctx, gint row, gchar *out)
+{
+    ChafaCanvas *canvas;
+    gint i, i_max;
+
+    canvas = ctx->canvas;
+    i = row * canvas->config.width;
+    i_max = (row + 1) * canvas->config.width;
+
+    if (row == 0
+        && canvas->config.canvas_mode != CHAFA_CANVAS_MODE_FGBG)
+    {
+        if (canvas->config.fg_only_enabled)
+            out = reset_fg (ctx, out);
+        else
+            out = reset_attributes (ctx, out);
+    }
+
+    switch (canvas->config.canvas_mode)
+    {
+        case CHAFA_CANVAS_MODE_TRUECOLOR:
+            out = emit_ansi_truecolor (ctx, out, i, i_max);
+            break;
+        case CHAFA_CANVAS_MODE_INDEXED_256:
+        case CHAFA_CANVAS_MODE_INDEXED_240:
+            out = emit_ansi_256 (ctx, out, i, i_max);
+            break;
+        case CHAFA_CANVAS_MODE_INDEXED_16:
+            out = emit_ansi_16 (ctx, out, i, i_max);
+            break;
+        case CHAFA_CANVAS_MODE_INDEXED_16_8:
+            out = emit_ansi_16_8 (ctx, out, i, i_max);
+            break;
+        case CHAFA_CANVAS_MODE_INDEXED_8:
+            out = emit_ansi_16 (ctx, out, i, i_max);
+            break;
+        case CHAFA_CANVAS_MODE_FGBG_BGFG:
+            out = emit_ansi_fgbg_bgfg (ctx, out, i, i_max);
+            break;
+        case CHAFA_CANVAS_MODE_FGBG:
+            out = emit_ansi_fgbg (ctx, out, i, i_max);
+            break;
+        case CHAFA_CANVAS_MODE_MAX:
+            g_assert_not_reached ();
+            break;
+    }
+
+    out = flush_chars (ctx, out);
+
+    /* Avoid control codes in FGBG mode. Don't reset attributes when BG
+     * is held, to preserve any BG color set previously. */
+    if (canvas->config.canvas_mode != CHAFA_CANVAS_MODE_FGBG)
+    {
+        if (canvas->config.fg_only_enabled)
+            out = reset_fg (ctx, out);
+        else
+            out = reset_attributes (ctx, out);
+    }
+
+    return out;
+}
+
 static GString *
 build_ansi_gstring (ChafaCanvas *canvas, ChafaTermInfo *ti)
 {
     GString *gs = g_string_new ("");
     PrintCtx ctx = { 0 };
-    gint i, i_max, i_step, i_next;
+    gint i;
 
     ctx.canvas = canvas;
     ctx.term_info = ti;
 
-    i = 0;
-    i_max = canvas->config.width * canvas->config.height;
-    i_step = canvas->config.width;
-
-    for ( ; i < i_max; i = i_next)
+    for (i = 0; i < canvas->config.height; i++)
     {
         gchar *out;
 
-        i_next = i + i_step;
-
-        prealloc_string (gs, i_step);
+        prealloc_string (gs, canvas->config.width);
         out = gs->str + gs->len;
 
-        /* Avoid control codes in FGBG mode. Don't reset attributes when BG
-         * is held, to preserve any BG color set previously. */
-        if (i == 0
-            && canvas->config.canvas_mode != CHAFA_CANVAS_MODE_FGBG
-            && !canvas->config.fg_only_enabled)
-        {
-            out = reset_attributes (&ctx, out);
-        }
+        out = build_ansi_row (&ctx, i, out);
 
-        switch (canvas->config.canvas_mode)
-        {
-            case CHAFA_CANVAS_MODE_TRUECOLOR:
-                out = emit_ansi_truecolor (&ctx, out, i, i_next);
-                break;
-            case CHAFA_CANVAS_MODE_INDEXED_256:
-            case CHAFA_CANVAS_MODE_INDEXED_240:
-                out = emit_ansi_256 (&ctx, out, i, i_next);
-                break;
-            case CHAFA_CANVAS_MODE_INDEXED_16:
-                out = emit_ansi_16 (&ctx, out, i, i_next);
-                break;
-            case CHAFA_CANVAS_MODE_INDEXED_16_8:
-                out = emit_ansi_16_8 (&ctx, out, i, i_next);
-                break;
-            case CHAFA_CANVAS_MODE_INDEXED_8:
-                out = emit_ansi_16 (&ctx, out, i, i_next);
-                break;
-            case CHAFA_CANVAS_MODE_FGBG_BGFG:
-                out = emit_ansi_fgbg_bgfg (&ctx, out, i, i_next);
-                break;
-            case CHAFA_CANVAS_MODE_FGBG:
-                out = emit_ansi_fgbg (&ctx, out, i, i_next);
-                break;
-            case CHAFA_CANVAS_MODE_MAX:
-                g_assert_not_reached ();
-                break;
-        }
-
-        out = flush_chars (&ctx, out);
-
-        /* Avoid control codes in FGBG mode. Don't reset attributes when BG
-         * is held, to preserve any BG color set previously. */
-        if (canvas->config.canvas_mode != CHAFA_CANVAS_MODE_FGBG
-            && !canvas->config.fg_only_enabled)
-        {
-            out = reset_attributes (&ctx, out);
-        }
-
-        /* Last line should not end in newline */
-        if (i_next < i_max)
+        /* Rows end in a newline, except for the last one. This allows for
+         * filling the display area while avoiding a blank bottom row. */
+        if (i < canvas->config.height - 1)
             *(out++) = '\n';
 
         *out = '\0';
@@ -739,6 +762,41 @@ build_ansi_gstring (ChafaCanvas *canvas, ChafaTermInfo *ti)
     return gs;
 }
 
+static void
+build_ansi_gstring_array (ChafaCanvas *canvas, ChafaTermInfo *ti,
+                          GString ***array_out, gint *array_len_out)
+{
+    PrintCtx ctx = { 0 };
+    GString **array;
+    gint i;
+
+    ctx.canvas = canvas;
+    ctx.term_info = ti;
+
+    array = g_new (GString *, canvas->config.height + 1);
+
+    for (i = 0; i < canvas->config.height; i++)
+    {
+        GString *gs = g_string_new ("");
+        gchar *out;
+
+        prealloc_string (gs, canvas->config.width);
+        out = gs->str + gs->len;
+
+        out = build_ansi_row (&ctx, i, out);
+        *out = '\0';
+
+        gs->len = out - gs->str;
+        array [i] = gs;
+    }
+
+    array [canvas->config.height] = NULL;
+    *array_out = array;
+
+    if (array_len_out)
+        *array_len_out = canvas->config.height;
+}
+
 GString *
 chafa_canvas_print_symbols (ChafaCanvas *canvas, ChafaTermInfo *ti)
 {
@@ -747,3 +805,14 @@ chafa_canvas_print_symbols (ChafaCanvas *canvas, ChafaTermInfo *ti)
 
     return build_ansi_gstring (canvas, ti);
 }
+
+void
+chafa_canvas_print_symbol_rows (ChafaCanvas *canvas, ChafaTermInfo *ti,
+                                GString ***array_out, gint *array_len_out)
+{
+    g_assert (canvas != NULL);
+    g_assert (ti != NULL);
+    g_assert (array_out != NULL);
+
+    build_ansi_gstring_array (canvas, ti, array_out, array_len_out);
+}
