@@ -1,8 +1,8 @@
 /* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
 
-/* Copyright (C) 2019-2022 Hans Petter Jansson
+/* Copyright (C) 2019-2023 Hans Petter Jansson
  *
- * This file is part of Chafa, a program that turns images into character art.
+ * This file is part of Chafa, a program that shows pictures on text terminals.
  *
  * Chafa is free software: you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published
@@ -22,6 +22,7 @@
 #include "smolscale/smolscale.h"
 #include "chafa.h"
 #include "internal/chafa-batch.h"
+#include "internal/chafa-math-util.h"
 #include "internal/chafa-private.h"
 
 typedef struct
@@ -60,17 +61,18 @@ gen_color_lut_rgba8 (guint32 *color_lut, ChafaColor col)
 }
 
 static void
-post_scale_row (guint32 *row_inout, int width, void *user_data)
+post_scale_row (gpointer row_inout, int width, void *user_data)
 {
     const DrawPixelsCtx *ctx = user_data;
-    guint32 *row_inout_end = row_inout + width;
+    guint32 *row_inout_u32 = row_inout;
+    guint32 *row_inout_end = row_inout_u32 + width;
 
     /* Composite on solid background color */
 
-    for ( ; row_inout < row_inout_end; row_inout++)
+    for ( ; row_inout_u32 < row_inout_end; row_inout_u32++)
     {
-        ChafaColor c = chafa_color8_fetch_from_rgba8 (row_inout);
-        *row_inout += ctx->bg_color_lut [c.ch [3]];
+        ChafaColor c = chafa_color8_fetch_from_rgba8 (row_inout_u32);
+        *row_inout_u32 += ctx->bg_color_lut [c.ch [3]];
     }
 }
 
@@ -400,7 +402,18 @@ chafa_indexed_image_new (gint width, gint height,
     indexed_image = g_new0 (ChafaIndexedImage, 1);
     indexed_image->width = width;
     indexed_image->height = height;
-    indexed_image->pixels = g_malloc (width * height);
+
+    indexed_image->pixels = g_try_malloc ((gsize) width * height);
+    if (!indexed_image->pixels)
+    {
+#if 0
+        g_warning ("ChafaIndexedImage: Out of memory allocating %ux%u pixels.",
+                   width, height);
+#endif
+
+        g_free (indexed_image);
+        return NULL;
+    }
 
     chafa_palette_copy (palette, &indexed_image->palette);
     chafa_palette_set_transparent_index (&indexed_image->palette, 255);
@@ -424,9 +437,14 @@ chafa_indexed_image_draw_pixels (ChafaIndexedImage *indexed_image,
                                  ChafaPixelType src_pixel_type,
                                  gconstpointer src_pixels,
                                  gint src_width, gint src_height, gint src_rowstride,
-                                 gint dest_width, gint dest_height)
+                                 gint dest_width, gint dest_height,
+                                 ChafaAlign halign, ChafaAlign valign,
+                                 ChafaTuck tuck)
 {
     DrawPixelsCtx ctx;
+    ChafaColor bg;
+    gint placement_x, placement_y;
+    gint placement_width, placement_height;
 
     g_return_if_fail (dest_width == indexed_image->width);
     g_return_if_fail (dest_height <= indexed_image->height);
@@ -444,23 +462,72 @@ chafa_indexed_image_draw_pixels (ChafaIndexedImage *indexed_image,
     ctx.dest_width = dest_width;
     ctx.dest_height = dest_height;
 
+#if 0
+    /* FIXME: Need a new smolscale compositing mode that preserves src
+     * alpha before this can be implemented */
+    bg = *chafa_palette_get_color (&indexed_image->palette,
+                                   CHAFA_COLOR_SPACE_RGB,
+                                   CHAFA_PALETTE_INDEX_BG);
+    bg.ch [3] = 0xff;
+#else
     gen_color_lut_rgba8 (ctx.bg_color_lut,
                          *chafa_palette_get_color (&indexed_image->palette,
                                                    CHAFA_COLOR_SPACE_RGB,
                                                    CHAFA_PALETTE_INDEX_BG));
+#endif
 
-    ctx.scaled_data = g_new (guint32, dest_width * dest_height);
-    ctx.scale_ctx = smol_scale_new_full ((SmolPixelType) src_pixel_type,
+    chafa_tuck_and_align (src_width, src_height,
+                          dest_width, dest_height,
+                          halign, valign,
+                          tuck,
+                          &placement_x, &placement_y,
+                          &placement_width, &placement_height);
+
+    /* FIXME: Save temp memory by sampling the image in strips. ChafaPalette
+     * will need a batch API for this. */
+    ctx.scaled_data = g_try_new (guint32, (gsize) dest_width * dest_height);
+    if (!ctx.scaled_data)
+    {
+#if 0
+        g_warning ("ChafaIndexedImage: Out of memory allocating %ux%u temp pixels.",
+                   dest_width, dest_height);
+#endif
+        return;
+    }
+
+    ctx.scale_ctx = smol_scale_new_full (/* Source */
                                          (const guint32 *) src_pixels,
+                                         (SmolPixelType) src_pixel_type,
                                          src_width,
                                          src_height,
                                          src_rowstride,
-                                         SMOL_PIXEL_RGBA8_PREMULTIPLIED,
+                                         /* Fill */
+#if 0
+                                         bg.ch,
+#else
                                          NULL,
+#endif
+                                         SMOL_PIXEL_RGBA8_UNASSOCIATED,
+                                         /* Destination */
+                                         NULL,
+                                         SMOL_PIXEL_RGBA8_PREMULTIPLIED,
                                          dest_width,
                                          dest_height,
                                          dest_width * sizeof (guint32),
+                                         /* Placement */
+                                         placement_x * SMOL_SUBPIXEL_MUL,
+                                         placement_y * SMOL_SUBPIXEL_MUL,
+                                         placement_width * SMOL_SUBPIXEL_MUL,
+                                         placement_height * SMOL_SUBPIXEL_MUL,
+                                         /* Extra args */
+                                         SMOL_COMPOSITE_SRC_CLEAR_DEST,
+#if 0
+                                         SMOL_NO_FLAGS,
+                                         NULL,
+#else
+                                         SMOL_DISABLE_SRGB_LINEARIZATION,
                                          post_scale_row,
+#endif
                                          &ctx);
 
     draw_pixels (&ctx);
