@@ -1,8 +1,8 @@
 /* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
 
-/* Copyright (C) 2021-2022 Hans Petter Jansson
+/* Copyright (C) 2021-2023 Hans Petter Jansson
  *
- * This file is part of Chafa, a program that turns images into character art.
+ * This file is part of Chafa, a program that shows pictures on text terminals.
  *
  * Chafa is free software: you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published
@@ -36,7 +36,9 @@ void chafa_iterm2_canvas_destroy (ChafaIterm2Canvas *iterm2_canvas);
 
 void chafa_iterm2_canvas_draw_all_pixels (ChafaIterm2Canvas *iterm2_canvas, ChafaPixelType src_pixel_type,
                                           gconstpointer src_pixels,
-                                          gint src_width, gint src_height, gint src_rowstride);
+                                          gint src_width, gint src_height, gint src_rowstride,
+                                          ChafaAlign halign, ChafaAlign valign,
+                                          ChafaTuck tuck);
 void chafa_iterm2_canvas_build_ansi (ChafaIterm2Canvas *iterm2_canvas, ChafaTermInfo *term_info, GString *out_str,
                                      gint width_cells, gint height_cells);
 
