@@ -1,8 +1,8 @@
 /* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
 
-/* Copyright (C) 2021-2022 Hans Petter Jansson
+/* Copyright (C) 2021-2023 Hans Petter Jansson
  *
- * This file is part of Chafa, a program that turns images into character art.
+ * This file is part of Chafa, a program that shows pictures on text terminals.
  *
  * Chafa is free software: you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published
@@ -40,6 +40,10 @@ chafa_format_dec_u8 (gchar *dest, guint8 n)
  * byte after the formatted ASCII decimal number (dest + 1..4). */
 gchar *chafa_format_dec_uint_0_to_9999 (char *dest, guint arg);
 
+/* Will overwrite 4 bytes starting at dest. Returns a pointer to the first
+ * byte after the formatted ASCII hexadecimal number (dest + 4). */
+gchar *chafa_format_dec_u16_hex (char *dest, guint16 arg);
+
 G_END_DECLS
 
 #endif /* __CHAFA_STRING_UTIL_H__ */
